#!/usr/bin/env perl
# sendmail.boot - "sendmail" boot-time setup script
# Revision: 070604

#---------------------------------------------------------------------
#                              overview
#---------------------------------------------------------------------

# "/etc/rc.d/rc.local" executes this script at boot time.  This script
# configures "sendmail".  Note: This script is  also  used  at "build"
# time.

# Note: The ".boot" filename extension is required. This extension en-
# sures that this script will be executed and not "sourced".

# Important:  Modify the settings defined in the  "program parameters"
# section as necessary.  These settings  must be  consistent  with the
# local "sendmail's" build-time configuration files [most importantly,
# "sendmail.mc" and "site.config.m4"].

#---------------------------------------------------------------------
#                        standard module setup
#---------------------------------------------------------------------

require 5.8.0;
use strict;
use Carp;
use warnings;
                                # Trap warnings
$SIG {__WARN__} = sub { die @_; };

#---------------------------------------------------------------------
#                           basic constants
#---------------------------------------------------------------------

use constant ZERO  => 0;        # Zero
use constant ONE   => 1;        # One

use constant FALSE => 0;        # Boolean FALSE
use constant TRUE  => 1;        # Boolean TRUE

#---------------------------------------------------------------------
#                         program parameters
#---------------------------------------------------------------------

#  $IEBASE specifies an internal-error message prefix string.

my $IEBASE = 'Internal error: sendmail.boot';

#---------------------------------------------------------------------

# The "$DIR_..." parameters specify absolute paths for  various "send-
# mail" directories:
#
#  $DIR_ETC_MAIL     = database directory [usually /etc/mail]
#  $DIR_VAR_MAIL     = mailbox  directory [usually /var/mail]

#  $DIR_CLIENTMQUEUE = "clientmqueue" directory
#                      [usually /var/spool/clientmqueue]

#  $DIR_MQUEUE       = "mqueue" directory
#                      [usually /var/spool/mqueue]

#  $DIR_HOSTSTAT     = "hoststat" directory
#                      [usually /var/spool/mqueue/.hoststat]

my $DIR_ETC_MAIL     = '/etc/mail';
my $DIR_VAR_MAIL     = '/var/mail';
my $DIR_CLIENTMQUEUE = '/var/spool/clientmqueue';
my $DIR_MQUEUE       = '/var/spool/mqueue';
my $DIR_HOSTSTAT     = "$DIR_MQUEUE/.hoststat";

#---------------------------------------------------------------------

# The "$EM_SRC_..." parameters specify  absolute paths for the various
# "sendmail" database source files.

my $EM_SRC_ACCESS   = "$DIR_ETC_MAIL/access";
my $EM_SRC_ALIASES  = "$DIR_ETC_MAIL/aliases";
my $EM_SRC_GENERICS = "$DIR_ETC_MAIL/generics";
my $EM_SRC_MAILER   = "$DIR_ETC_MAIL/mailer";
my $EM_SRC_VIRTUSER = "$DIR_ETC_MAIL/virtuser";

#---------------------------------------------------------------------

# The "$EM_TXT_..." parameters specify  absolute paths for some "send-
# mail" text-format configuration files.  "sendmail" reads these files
# directly;  i.e., we don't need to generate corresponding ".db" data-
# base files.

my $EM_TXT_GENERICS_DOMAINS = "$DIR_ETC_MAIL/generics-domains";
my $EM_TXT_LOCAL_HOST_NAMES = "$DIR_ETC_MAIL/local-host-names";

#---------------------------------------------------------------------

#  $DIR_BIN_SENDMAIL specifies an absolute path for  a standard  "bin"
# directory  that  contains the  "sendmail" executable  [or a symbolic
# link to the executable].  Usually,  this parameter should  be set to
# '/usr/bin'.

my $DIR_BIN_SENDMAIL = '/usr/bin';

#---------------------------------------------------------------------

#  $PROG_SENDMAIL specifies an absolute path for the "sendmail" execu-
# table [or a symbolic link to the executable].

my $PROG_SENDMAIL = "$DIR_BIN_SENDMAIL/sendmail";

#---------------------------------------------------------------------

#  $SENDMAIL_CYCLE_MINUTES specifies the number of minutes that "send-
# mail" should wait between outbound-mail cycles. Note: This parameter
# usually doesn't affect mail sent  from the local system  to the same
# system.

# If you're running a mail server for a large network, you may wish to
# increase this parameter to 15 or 30. Otherwise, values of one to ten
# are probably O.K.

my $SENDMAIL_CYCLE_MINUTES = 5;

#---------------------------------------------------------------------

#  $SENDMAIL_USRGRP specifies a special  user [and group] name used by
# "sendmail".  This parameter must be consistent with the value of the
# "sendmail" compile-time setting "confRUN_AS_USER".

# This  script assumes that "confRUN_AS_USER" specified identical user
# and group names [for example, "smmsp:smmsp"]. You can set $SENDMAIL_
# USRGRP equal to  the original value of  "confRUN_AS_USER"  or to the
# [single] name  used. If you use the value of "confRUN_AS_USER",  the
# user and group components  must be  identical,  or an internal-error
# message will be produced.

my $SENDMAIL_USRGRP = 'smmsp:smmsp';

#---------------------------------------------------------------------

                                # Adjust $SENDMAIL_USRGRP setting
if ($SENDMAIL_USRGRP =~ m@^(\w+):(\w+)\z@)
{
    die "$IEBASE #01 Invalid SENDMAIL_USRGRP setting\n" if $1 ne $2;
    $SENDMAIL_USRGRP = $1;
}
                                # Check the setting
if ($SENDMAIL_USRGRP !~ m@^\w+\z@)
{                               # Internal error
    die "$IEBASE #02 Invalid SENDMAIL_USRGRP setting\n";
}

#---------------------------------------------------------------------

#  $EM_MAKEFILE specifies an absolute path for the "sendmail" database
# "Makefile".

#  $PROCESS_SHUTDOWN_TIMEOUT specifies the  maximum number  of seconds
# to be allowed for a process shutdown after SIGTERM.

my $EM_MAKEFILE              = "$DIR_ETC_MAIL/Makefile";
my $PROCESS_SHUTDOWN_TIMEOUT = 5;

#---------------------------------------------------------------------
#                      directory-setup commands
#---------------------------------------------------------------------

#  $CMD_DIR_SETUP  defines a set of  shell-level  directory-setup com-
# mands.

#---------------------------------------------------------------------

my $CMD_DIR_SETUP = << "END";   # "END" must be double-quoted here
mkdir -p                        $DIR_ETC_MAIL
chown -R $SENDMAIL_USRGRP.root  $DIR_ETC_MAIL
chmod 755                       $DIR_ETC_MAIL
chown root.$SENDMAIL_USRGRP     $DIR_ETC_MAIL
chmod -R o-wrx                  $DIR_ETC_MAIL

mkdir -p        $DIR_VAR_MAIL
chown root.mail $DIR_VAR_MAIL
chmod 1777      $DIR_VAR_MAIL

mkdir -p                        $DIR_CLIENTMQUEUE
chown -R root.$SENDMAIL_USRGRP  $DIR_CLIENTMQUEUE
chmod 770                       $DIR_CLIENTMQUEUE

mkdir -p                        $DIR_MQUEUE
chown -R root.$SENDMAIL_USRGRP  $DIR_MQUEUE
chmod 770                       $DIR_MQUEUE

mkdir -p                        $DIR_HOSTSTAT
chown root.root                 $DIR_HOSTSTAT
chmod 755                       $DIR_HOSTSTAT
END

#---------------------------------------------------------------------
#                  template for "access" source file
#---------------------------------------------------------------------

#  $EM_SRC_ACCESS_DATA  defines a template that's used  to create  the
# initial "sendmail access" source file.

# Note:  Network administrators may modify  the file  after  it's been
# created. If the file is modified, the following  command should  up-
# date things appropriately  [modify  the path  shown here,  if neces-
# sary]:
#
#     make -f /etc/mail/Makefile

#---------------------------------------------------------------------

# Future change: Should hostname be included here?

my $EM_SRC_ACCESS_DATA = << 'END';
localhost.localdomain   RELAY
localhost               RELAY
127.0.0.1               RELAY
END

#---------------------------------------------------------------------
#                 template for "aliases" source file
#---------------------------------------------------------------------

#  $EM_SRC_ALIASES_DATA  defines a template that's used  to create the
# initial "sendmail aliases" source file.

# Note:  Network administrators may modify  the file  after  it's been
# created. If the file is modified, the following  command should  up-
# date things appropriately  [modify  the path  shown here,  if neces-
# sary]:
#
#     make -f /etc/mail/Makefile

#---------------------------------------------------------------------

my $EM_SRC_ALIASES_DATA = << 'END';
# __META_ETC_MAIL__/aliases - "sendmail" aliases source file

# Note:  If you modify this file,  you'll  need to execute the follow-
# ing command:
#
#     make -f __META_ETC_MAIL__/Makefile
#
# This command will rebuild the "sendmail" aliases database using this
# file as input.

#---------------------------------------------------------------------

# Basic system aliases. These MUST be present.
MAILER-DAEMON: postmaster
postmaster: root

# Miscellaneous aliases.
abuse:      root
admin:      root
manager:    root
END

#---------------------------------------------------------------------

                                # Translate "META" parameter(s)
$EM_SRC_ALIASES_DATA =~ s@_+META_ETC_MAIL_+@$DIR_ETC_MAIL@g;

                                # Consistency check
die "$IEBASE #03\n" if $EM_SRC_ALIASES_DATA =~ m@META@;

#---------------------------------------------------------------------
#                 template for "generics" source file
#---------------------------------------------------------------------

#  $EM_SRC_GENERICS_DATA  defines a template that's used to create the
# initial "sendmail generics" source file.

# Note:  Network administrators may modify  the file  after  it's been
# created. If the file is modified, the following  command should  up-
# date things appropriately  [modify  the path  shown here,  if neces-
# sary]:
#
#     make -f /etc/mail/Makefile

#---------------------------------------------------------------------

my $EM_SRC_GENERICS_DATA = << 'END';
john    john@put.your.domain.here
mary    mary@put.your.domain.here
END

#---------------------------------------------------------------------
#               template for "mailer-table" source file
#---------------------------------------------------------------------

#  $EM_SRC_MAILER_DATA  defines a  template that's  used  to create the
# initial "sendmail mailer-table" source file.

# Note:  Network administrators may modify  the file  after  it's been
# created. If the file is modified, the following  command should  up-
# date things appropriately  [modify  the path  shown here,  if neces-
# sary]:
#
#     make -f /etc/mail/Makefile

#---------------------------------------------------------------------

my $EM_SRC_MAILER_DATA = << 'END';
nosuchdomain.unk    error:450 This is a sample "error" entry
END

#---------------------------------------------------------------------
#                 template for "virtuser" source file
#---------------------------------------------------------------------

#  $EM_SRC_VIRTUSER_DATA  defines a template that's used to create the
# initial "sendmail virtuser" source file.

# Note:  Network administrators may modify  the file  after  it's been
# created. If the file is modified, the following  command should  up-
# date things appropriately  [modify  the path  shown here,  if neces-
# sary]:
#
#     make -f /etc/mail/Makefile

#---------------------------------------------------------------------

my $EM_SRC_VIRTUSER_DATA = << 'END';
admin@local.system        john
dave@some-domain.com      frodo@another-domain.com
@yet-another-domain.org   mary
END

#---------------------------------------------------------------------
#             template for "sendmail" database "Makefile"
#---------------------------------------------------------------------

#  $EM_MAKEFILE_DATA  defines a  template that's  used  to create  the
# "sendmail" database "Makefile".

#---------------------------------------------------------------------

# This code creates the first part of $EM_MAKEFILE_DATA. Some database
# creation rules are omitted; these are added in the next section.

my $EM_MAKEFILE_DATA = << 'END';
# __META_ETC_MAIL__/Makefile - "sendmail" database "Makefile"

EM          = /etc/mail
MAKEMAP     = /usr/sbin/makemap
NEWALIASES  = /usr/sbin/newaliases

EM_ACCESS   = $(EM)/access
EM_ALIASES  = $(EM)/aliases
EM_GENERICS = $(EM)/generics
EM_MAILER   = $(EM)/mailer
EM_VIRTUSER = $(EM)/virtuser

ALL_DB_FILES = \
    $(EM_ACCESS).db    \
    $(EM_ALIASES).db   \
    $(EM_GENERICS).db  \
    $(EM_MAILER).db    \
    $(EM_VIRTUSER).db

all: $(ALL_DB_FILES)

clean:
__META_TAB__rm -f $(ALL_DB_FILES)

$(EM_ALIASES).db: $(NEWALIASES) $(EM_ALIASES)
__META_TAB__rm -f $(EM_ALIASES).db
__META_TAB__touch $(EM_ALIASES).db
__META_TAB__chown smmsp.root $(EM_ALIASES).db
__META_TAB__chmod 644 $(EM_ALIASES).db
__META_TAB__$(NEWALIASES) -v
__META_TAB__chmod 444 $(EM_ALIASES).db
__META_TAB__chown __META_USRGRP__.root $(EM_ALIASES).db
END

#---------------------------------------------------------------------

# This code  adds the remaining  database creation rules to  $EM_MAKE-
# FILE_DATA.

for my $keyword (qw (ACCESS GENERICS MAILER VIRTUSER))
{
    my $str = << 'END';         # 'END' must be single-quoted here
$(EM_KEYWORD).db: $(MAKEMAP) $(EM_KEYWORD)
__META_TAB__rm -f $(EM_KEYWORD).db
__META_TAB__chmod a+w $(EM_KEYWORD)*
__META_TAB__$(MAKEMAP) hash $(EM_KEYWORD) < $(EM_KEYWORD)
__META_TAB__chmod a-w $(EM_KEYWORD)*
__META_TAB__chown __META_USRGRP__.root $(EM_KEYWORD).db
END
    $str =~ s@KEYWORD@$keyword@g;
    $EM_MAKEFILE_DATA .= "\n";
    $EM_MAKEFILE_DATA .= $str;
}

#---------------------------------------------------------------------

# This code translates a few meta-symbols to create the  final version
# of the "Makefile" text.

                                # Translate "META" parameters
$EM_MAKEFILE_DATA =~ s@_+META_ETC_MAIL_+@$DIR_ETC_MAIL@g;
$EM_MAKEFILE_DATA =~ s@_+META_TAB_+@\t@g;
$EM_MAKEFILE_DATA =~ s@_+META_USRGRP_+@$SENDMAIL_USRGRP@g;

                                # Consistency check
die "$IEBASE #04\n" if $EM_MAKEFILE_DATA =~ m@META@;

#---------------------------------------------------------------------
#                          utility routines
#---------------------------------------------------------------------

# ReadFile - Reads an entire file.

# Usage:
#
# my $data = &ReadFile (-path => $pathname);
#
# string $pathname  - Pathname for input file [may be absolute or rel-
# ative.

#---------------------------------------------------------------------

sub ReadFile
{
    my $rtn = 'ReadFile';       # Routine name
    my (%args)    = @_;         # Argument list
    my $SaveSlash = $/;         # Original record delimiter
    my $data;                   # Data buffer

#---------------------------------------------------------------------
# Initial setup.

    my $path = $args {-path};   # Pathname
                                # Consistency check
    die "Internal error\n" if !defined $path;

#---------------------------------------------------------------------
# Read entire configuration file.

    if (!open (IFD, "<$path"))
    {
        die "Error: $rtn: Can't open file: $!\n$path\n";
    }

    undef $/;                   # Set entire-file mode
    binmode (IFD);              # Set raw-data mode
    $data = <IFD>;              # Read entire file
                                # Safety measure
    $data = "" if !defined ($data);
    close (IFD);                # Close the file
    $/ = $SaveSlash;            # Restore record separator

    $data;                      # Contents of file
}

#---------------------------------------------------------------------
#                            main routine
#---------------------------------------------------------------------

sub Main
{
    my $cmd;                    # Shell-level command(s)
    my $str;                    # Scratch
    my $ShortName;              # Ditto

#---------------------------------------------------------------------
# Print a status message.

    if (-f $PROG_SENDMAIL)      # Does "sendmail" executable exist?
    {                           # Yes
        print "Setting up sendmail\n";
    }
    else
    {                           # No
        print "Setting up sendmail-related files\n";
    }

#---------------------------------------------------------------------
# Shut down existing "sendmail" process [if any].

    my $Timeout = FALSE;        # "Timeout" flag

    eval
    {                           # Handle timeouts
        local ($SIG{ALRM}) = sub { $Timeout = TRUE; die(); };

                                # Start alarm clock
        alarm ($PROCESS_SHUTDOWN_TIMEOUT);

                                # Try to shut process down cleanly
        system "killall -qw --signal SIGTERM sendmail 2>/dev/null";

        alarm (ZERO);           # Reset alarm clock
    };

    if ($Timeout)               # Did shutdown time-out?
    {                           # Yes - Take stronger measures
        print STDERR << 'END';
Warning: Couldn't shut "sendmail" down cleanly. Using SIGKILL to force
shutdown.
END
        system "killall --signal SIGKILL sendmail";
        sleep 2;
    }

#---------------------------------------------------------------------
# Read group/user data.

    my $GroupData = &ReadFile (-path => '/etc/group' );
    my $UserData  = &ReadFile (-path => '/etc/passwd');

#---------------------------------------------------------------------
# Create special group, if necessary.

    if ($GroupData !~ m@(^|\n)$SENDMAIL_USRGRP:@)
    {
        $cmd = "groupadd -g 26 $SENDMAIL_USRGRP";
        print "$cmd\n";
        system $cmd;
    }

#---------------------------------------------------------------------
# Create "mail" group, if necessary.

    if ($GroupData !~ m@(^|\n)mail:@)
    {
        $cmd = "groupadd -g 34 mail";
        print "$cmd\n";
        system $cmd;
    }

#---------------------------------------------------------------------
# Create special user, if necessary.

    if ($UserData !~ m@(^|\n)$SENDMAIL_USRGRP:@)
    {
        $cmd = << "END";        # "END" must be double-quoted here
useradd -c "sendmail daemon"
    -g $SENDMAIL_USRGRP -G mail -d /dev/null -s /bin/false
    -u 26 $SENDMAIL_USRGRP
END
        $cmd =~ s@\s+@ @gs;
        $cmd =~ s@\s+\z@@s;
        print "$cmd\n";
        system $cmd;
    }

#---------------------------------------------------------------------
# Create "mail" user, if necessary.

    if ($UserData !~ m@(^|\n)mail:@)
    {
$cmd = << 'END';
useradd -g mail -d /dev/null -s /bin/false -u 34 mail
END
        $cmd =~ s@\s+@ @gs;
        $cmd =~ s@\s+\z@@s;
        print "$cmd\n";
        system $cmd;
    }

#---------------------------------------------------------------------
# Set up various directories.

    system $CMD_DIR_SETUP;

#---------------------------------------------------------------------
# Set up "access" database source file.

    if (!-f $EM_SRC_ACCESS)     # Need to create file?
    {                           # Yes
        open (OFD, ">$EM_SRC_ACCESS") ||
            die "Error: Can't create file: $!\n$EM_SRC_ACCESS\n";
        print OFD $EM_SRC_ACCESS_DATA;
        close OFD;
    }

    system "chown $SENDMAIL_USRGRP.root $EM_SRC_ACCESS";
    system "chmod 444 $EM_SRC_ACCESS";

#---------------------------------------------------------------------
# Set up "aliases" database source file.

    if (!-f $EM_SRC_ALIASES)    # Need to create file?
    {                           # Yes
        open (OFD, ">$EM_SRC_ALIASES") ||
            die "Error: Can't create file: $!\n$EM_SRC_ALIASES\n";
        print OFD $EM_SRC_ALIASES_DATA;
        close OFD;
    }

    system "chown $SENDMAIL_USRGRP.root $EM_SRC_ALIASES";
    system "chmod 444 $EM_SRC_ALIASES";

#---------------------------------------------------------------------
# Set up "generics" database source file.

    if (!-f $EM_SRC_GENERICS)   # Need to create file?
    {                           # Yes
        open (OFD, ">$EM_SRC_GENERICS") ||
            die "Error: Can't create file: $!\n$EM_SRC_GENERICS\n";
        print OFD $EM_SRC_GENERICS_DATA;
        close OFD;
    }

    system "chown $SENDMAIL_USRGRP.root $EM_SRC_GENERICS";
    system "chmod 444 $EM_SRC_GENERICS";

#---------------------------------------------------------------------
# Set up "mailer-table" database source file.

    if (!-f $EM_SRC_MAILER)     # Need to create file?
    {                           # Yes
        open (OFD, ">$EM_SRC_MAILER") ||
            die "Error: Can't create file: $!\n$EM_SRC_MAILER\n";
        print OFD $EM_SRC_MAILER_DATA;
        close OFD;
    }

    system "chown $SENDMAIL_USRGRP.root $EM_SRC_MAILER";
    system "chmod 444 $EM_SRC_MAILER";

#---------------------------------------------------------------------
# Set up "virtuser" database source file.

    if (!-f $EM_SRC_VIRTUSER)   # Need to create file?
    {                           # Yes
        open (OFD, ">$EM_SRC_VIRTUSER") ||
            die "Error: Can't create file: $!\n$EM_SRC_VIRTUSER\n";
        print OFD $EM_SRC_VIRTUSER_DATA;
        close OFD;
    }

    system "chown $SENDMAIL_USRGRP.root $EM_SRC_VIRTUSER";
    system "chmod 444 $EM_SRC_VIRTUSER";

#---------------------------------------------------------------------
# Set up "sendmail" database "Makefile".

    if (!-f $EM_MAKEFILE)       # Need to create file?
    {                           # yes
        open (OFD, ">$EM_MAKEFILE") ||
            die "Error: Can't create file: $!\n$EM_MAKEFILE\n";
        print OFD $EM_MAKEFILE_DATA;
        close OFD;
    }

    system "chown $SENDMAIL_USRGRP.root $EM_MAKEFILE";
    system "chmod 444 $EM_MAKEFILE";

#---------------------------------------------------------------------
# Set up "generics-domains" configuration file.

    $ShortName = $EM_TXT_GENERICS_DOMAINS;

    if (!-f $ShortName)
    {
        system "touch $ShortName";
    }

    system "chown $SENDMAIL_USRGRP.root $ShortName";
    system "chmod 444 $ShortName";

#---------------------------------------------------------------------
# Set up "local-host-names" configuration file.

    $ShortName = $EM_TXT_LOCAL_HOST_NAMES;

    if (!-f $ShortName)
    {
        open (OFD, ">$ShortName") ||
            die "Error: Can't create file: $!\n$ShortName\n";

        print OFD << 'END';
localhost
localhost.localdomain
END
        close OFD;

# Future change:  Add the  FQDN only  if it exists and  it's different
# than the hostname.

        system "/bin/hostname >> $ShortName";
        system "/bin/hostname --fqdn >> $ShortName";
    }

# Future change:  This program should chown/chmod files if and only if
# it's necessary.

    system "chown $SENDMAIL_USRGRP.root $ShortName";
    system "chmod 444 $ShortName";

#---------------------------------------------------------------------
# Misc. adjustments.

    for my $basename (qw (sendmail submit))
    {
        $str = "$DIR_ETC_MAIL/$basename.cf";
        system "chmod 444 $str" if -f $str;
    }

#---------------------------------------------------------------------
# Wrap it up.

    if ((-f $PROG_SENDMAIL) && (-x $PROG_SENDMAIL))
    {
        system "chown root.smmsp $PROG_SENDMAIL";
        system "chmod 555        $PROG_SENDMAIL";
        system "chmod g+s        $PROG_SENDMAIL";

        my $OUTFILE = "/tmp/sendmail-boot-tmp1-$>-$$";
        my $ERRFILE = "/tmp/sendmail-boot-tmp2-$>-$$";

        chdir ($DIR_ETC_MAIL);
        unlink $ERRFILE;
        die "$IEBASE #05\n" if -e $ERRFILE;
        system "make >$OUTFILE 2>$ERRFILE";
        die "$IEBASE #06\n" if !-T $ERRFILE;

        $str =  `cat $OUTFILE $ERRFILE 2>/dev/null`;
        $str =  "" if !defined $str;
        $str =~ s@\s+\z@@s;

        my $is_error = (-z $ERRFILE) ? FALSE : TRUE;
        unlink $OUTFILE;
        unlink $ERRFILE;

        if ($is_error)
        {
            print "$str\n";
            exit 1;
        }

        system "$PROG_SENDMAIL -bd -q${SENDMAIL_CYCLE_MINUTES}m &";
    }

    undef;
}

#---------------------------------------------------------------------
#                            main program
#---------------------------------------------------------------------

&Main();                        # Call the main routine
exit ZERO;                      # Normal exit
